-- CXB3010.A
--
--                             Grant of Unlimited Rights
--
--     Under contracts F33600-87-D-0337, F33600-84-D-0280, MDA903-79-C-0687,
--     F08630-91-C-0015, and DCA100-97-D-0025, the U.S. Government obtained
--     unlimited rights in the software and documentation contained herein.
--     Unlimited rights are defined in DFAR 252.227-7013(a)(19).  By making
--     this public release, the Government intends to confer upon all
--     recipients unlimited rights  equal to those held by the Government.
--     These rights include rights to use, duplicate, release or disclose the
--     released technical data and computer software in whole or in part, in
--     any manner and for any purpose whatsoever, and to have or permit others
--     to do so.
--
--                                    DISCLAIMER
--
--     ALL MATERIALS OR INFORMATION HEREIN RELEASED, MADE AVAILABLE OR
--     DISCLOSED ARE AS IS.  THE GOVERNMENT MAKES NO EXPRESS OR IMPLIED
--     WARRANTY AS TO ANY MATTER WHATSOEVER, INCLUDING THE CONDITIONS OF THE
--     SOFTWARE, DOCUMENTATION OR OTHER INFORMATION RELEASED, MADE AVAILABLE
--     OR DISCLOSED, OR THE OWNERSHIP, MERCHANTABILITY, OR FITNESS FOR A
--     PARTICULAR PURPOSE OF SAID MATERIAL.
--*
--
-- OBJECTIVE:
--      Check that the Procedure Free resets the parameter Item to
--      Null_Ptr.  Check that Free has no effect if Item is Null_Ptr.
--
--      Check that the version of Function Value with a chars_ptr parameter
--      returning a char_array result returns the prefix of an array of
--      chars.
--
--      Check that the version of Function Value with a chars_ptr parameter
--      and a size_t parameter returning a char_array result returns
--      the shorter of:
--        1) the first size_t number of characters, or
--        2) the characters up to and including the first nul.
--
--      Check that both of the above versions of Function Value propagate
--      Dereference_Error if the Item parameter is Null_Ptr.
--
-- TEST DESCRIPTION:
--      This test validates the Procedure Free and two versions of Function
--      Value.  A variety of char_array and char_ptr values are provided as
--      input, and results are compared for both length and content.
--
--      This test assumes that the following characters are all included
--      in the implementation defined type Interfaces.C.char:
--      ' ', 'a'..'z', and 'A'..'Z'.
--
-- APPLICABILITY CRITERIA:
--      This test is applicable to all implementations that provide
--      package Interfaces.C.Strings.  If an implementation provides
--      package Interfaces.C.Strings, this test must compile, execute,
--      and report "PASSED".
--
--
-- CHANGE HISTORY:
--      27 Sep 95   SAIC    Initial prerelease version.
--      13 May 96   SAIC    Incorporated reviewer comments for ACVC 2.1.
--      26 Oct 96   SAIC    Incorporated reviewer comments.
--      01 DEC 97   EDS     Replicate line 199 at line 256, to ensure that
--                          TC_chars_ptr has a valid pointer.
--      08 JUL 99   RLB     Added a test case to check that Value raises
--                          Constraint_Error when Length = 0. (From Technical
--                          Corrigendum 1).
--      25 JAN 01   RLB     Repaired previous test case to avoid raising
--                          Constraint_Error in test case code.
--      26 JAN 01   RLB     Added an Ident_Int to the test case to prevent
--                          optimization.

--!

with Report;
with Interfaces.C.Strings;                                    -- N/A => ERROR

procedure Cxb3010 is
begin

   Report.Test
     ("CXB3010",
      "Check that Procedure Free and versions of " &
      "Function Value produce correct results");

   Test_Block :
   declare

      package Ic renames Interfaces.C;
      package Ics renames Interfaces.C.Strings;

      use type Ic.Char_Array;
      use type Ic.Size_T;
      use type Ics.Chars_Ptr;
      use type Ic.Char;

      Null_Char_Array_Access : constant Ics.Char_Array_Access := null;

      Tc_String_1     : constant String          := "Nonul";
      Tc_String_2     : constant String          := "AbCdE";
      Tc_Blank_String : constant String (1 .. 5) := (others => ' ');

      -- The initialization of the following char_array objects includes the
      -- appending of a terminating nul char, in order to prevent the erroneous
      -- execution of Function Value.

      Tc_Char_Array       : Ic.Char_Array := Ic.To_C (Tc_Blank_String, True);
      Tc_Char_Array_1 : constant Ic.Char_Array := Ic.To_C (Tc_String_1, True);
      Tc_Char_Array_2 : constant Ic.Char_Array := Ic.To_C (Tc_String_2, True);
      Tc_Blank_Char_Array : constant Ic.Char_Array :=
        Ic.To_C (Tc_Blank_String, True);

      -- This chars_ptr is initialized via the use of New_Chars_Array to avoid
      -- erroneous execution of procedure Free.
      Tc_Chars_Ptr : Ics.Chars_Ptr := Ics.New_Char_Array (Tc_Blank_Char_Array);

   begin

      -- Check that the Procedure Free resets the parameter Item to Null_Ptr.

      if Tc_Chars_Ptr = Ics.Null_Ptr then
         Report.Failed
           ("TC_chars_ptr is currently null; it should not be " &
            "null since it was given default initialization");
      end if;

      Ics.Free (Tc_Chars_Ptr);

      if Tc_Chars_Ptr /= Ics.Null_Ptr then
         Report.Failed
           ("TC_chars_ptr was not set to Null_Ptr by " & "Procedure Free");
      end if;

      -- Check that Free has no effect if Item is Null_Ptr.

      begin
         Tc_Chars_Ptr := Ics.Null_Ptr;  -- Ensure pointer is null.
         Ics.Free (Tc_Chars_Ptr);
         if Tc_Chars_Ptr /= Ics.Null_Ptr then
            Report.Failed
              ("TC_chars_ptr was set to a non-Null_Ptr value " &
               "by Procedure Free.  It was provided as a null " &
               "parameter to Free, and there should have been " &
               "no effect from a call to Procedure Free");
         end if;
      exception
         when others =>
            Report.Failed
              ("Unexpected exception raised by Procedure Free " &
               "when parameter Item is Null_Ptr");
      end;

      -- Check that the version of Function Value with a chars_ptr parameter
      -- that returns a char_array result returns an array of chars (up to and
      -- including the first nul).

      Tc_Chars_Ptr  := Ics.New_Char_Array (Tc_Char_Array_1);
      Tc_Char_Array := Ics.Value (Item => Tc_Chars_Ptr);

      if Tc_Char_Array /= Tc_Char_Array_1 or
        Ic.To_Ada (Tc_Char_Array, True) /= Ic.To_Ada (Tc_Char_Array_1) then
         Report.Failed ("Incorrect result from Function Value - 1");
      end if;

      Tc_Chars_Ptr  := Ics.New_Char_Array (Tc_Char_Array_2);
      Tc_Char_Array := Ics.Value (Item => Tc_Chars_Ptr);

      if Tc_Char_Array /= Tc_Char_Array_2 or
        Ic.To_Ada (Tc_Char_Array, True) /= Ic.To_Ada (Tc_Char_Array_2) then
         Report.Failed ("Incorrect result from Function Value - 2");
      end if;

      if Ics.Value (Item => Ics.New_String ("A little longer string")) /=
        Ic.To_C ("A little longer string") then
         Report.Failed ("Incorrect result from Function Value - 3");
      end if;

      -- Check that the version of Function Value with a chars_ptr parameter
      -- and a size_t parameter that returns a char_array result returns the
      -- shorter of:
      --   1) the first size_t number of characters, or
      --   2) the characters up to and including the first nul.

      -- Case 1: the first size_t number of characters (less than the
      --         total length).

      begin
         Tc_Chars_Ptr           := Ics.New_Char_Array (Tc_Char_Array_1);
         Tc_Char_Array (0 .. 2) :=
           Ics.Value (Item => Tc_Chars_Ptr, Length => 3);

         if Tc_Char_Array (0 .. 2) /= Tc_Char_Array_1 (0 .. 2) then
            Report.Failed
              ("Incorrect result from Function Value with Length " &
               "parameter - 1");
         end if;
      exception
         when others =>
            Report.Failed ("Exception raised during Case 1 evaluation");
      end;

      -- Case 2: the characters up to and including the first nul.

      Tc_Chars_Ptr := Ics.New_Char_Array (Tc_Char_Array_2);

      -- The length supplied as a parameter exceeds the total length of
      -- TC_char_array_2. The result should be the entire TC_char_array_2
      -- including the terminating nul.

      Tc_Char_Array := Ics.Value (Item => Tc_Chars_Ptr, Length => 7);

      if Tc_Char_Array /= Tc_Char_Array_2 or
        Ic.To_Ada (Tc_Char_Array) /= Ic.To_Ada (Tc_Char_Array_2) or
        not (Ic.Is_Nul_Terminated (Tc_Char_Array)) then
         Report.Failed
           ("Incorrect result from Function Value with Length " &
            "parameter - 2");
      end if;

      -- Check that both of the above versions of Function Value propagate
      -- Dereference_Error if the Item parameter is Null_Ptr.

      declare

         -- Declare a dummy function to demonstrate one way that a chars_ptr
         -- variable could inadvertantly be set to Null_Ptr prior to a call
         -- to Value (below).
         function Freedom (Condition : Boolean := False;
            Ptr                      : Ics.Chars_Ptr) return Ics.Chars_Ptr
         is
            Pointer : Ics.Chars_Ptr := Ptr;
         begin
            if Condition then
               Ics.Free (Pointer);
            else
               null; -- An activity that doesn't set the chars_ptr value to
               -- Null_Ptr.
            end if;
            return Pointer;
         end Freedom;

      begin

         begin
            Tc_Char_Array := Ics.Value (Item => Freedom (True, Tc_Chars_Ptr));
            Report.Failed
              ("Function Value (without Length parameter) did not " &
               "raise Dereference_Error when provided a null Item " &
               "parameter input value");
            if Tc_Char_Array (0) = '6' then   -- Defeat optimization.
               Report.Comment ("Should never be printed");
            end if;
         exception
            when Ics.Dereference_Error =>
               null;  -- OK, expected exception.
            when others =>
               Report.Failed
                 ("Incorrect exception raised by Function Value " &
                  "with Item parameter, when the Item parameter " &
                  "is Null_Ptr");
         end;

         Tc_Chars_Ptr := Ics.New_Char_Array (Tc_Char_Array_2);
         begin
            Tc_Char_Array :=
              Ics.Value (Item => Freedom (True, Tc_Chars_Ptr), Length => 4);
            Report.Failed
              ("Function Value (with Length parameter) did not " &
               "raise Dereference_Error when provided a null Item " &
               "parameter input value");
            if Tc_Char_Array (0) = '6' then   -- Defeat optimization.
               Report.Comment ("Should never be printed");
            end if;
         exception
            when Ics.Dereference_Error =>
               null;  -- OK, expected exception.
            when others =>
               Report.Failed
                 ("Incorrect exception raised by Function Value " &
                  "with both Item and Length parameters, when " &
                  "the Item parameter is Null_Ptr");
         end;
      end;

      -- Check that Function Value with two parameters propagates
      -- Constraint_Error if Length is 0.

      begin
         Tc_Chars_Ptr := Ics.New_Char_Array (Tc_Char_Array_1);
         declare
            Tc : Ic.Char_Array :=
              Ics.Value
                (Item   => Tc_Chars_Ptr,
                 Length => Ic.Size_T (Report.Ident_Int (0)));
         begin
            Report.Failed
              ("Function Value (with Length parameter) did not " &
               "raise Constraint_Error when Length = 0");
            if Tc'Length <= Tc_Char_Array'Length then
               Tc_Char_Array (1 .. Tc'Length) :=
                 Tc; -- Block optimization of TC.
            end if;
         end;

         Report.Failed
           ("Function Value (with Length parameter) did not " &
            "raise Constraint_Error when Length = 0");
      exception
         when Constraint_Error =>
            null;  -- OK, expected exception.
         when others =>
            Report.Failed
              ("Incorrect exception raised by Function Value " &
               "with both Item and Length parameters, when " & "Length = 0");
      end;

   exception
      when others =>
         Report.Failed ("Exception raised in Test_Block");
   end Test_Block;

   Report.Result;

end Cxb3010;
